'''
intervaltree-bio: Interval tree convenience classes for genomic data.

Test module. Meant for use with ``py.test``.
To run all tests simply invoke ``python setup.py test`` from the package root.

Copyright 2014 Konstantin Tretyakov

Licensed under MIT.
'''
